""" powermon / protocols / pi18.py """
import logging

from powermon.commands.command import CommandType
from powermon.commands.command_definition import CommandDefinition
from powermon.commands.reading_definition import ReadingType, ResponseType
from powermon.commands.result import ResultType
from powermon.errors import CommandDefinitionMissing
from powermon.ports.porttype import PortType
from powermon.protocols.abstractprotocol import AbstractProtocol
from powermon.protocols.helpers import crc_pi30 as crc
from powermon.protocols.pi30 import BATTERY_TYPE_LIST, OUTPUT_MODE_LIST

log = logging.getLogger("pi18")

SETTER_COMMANDS = {}


QUERY_COMMANDS = {
    "PI": {
        "name": "PI",
        "command_type": CommandType.PI18_QUERY,
        "description": "Protocol ID inquiry",
        "help": " -- queries the protocol ID",
        "result_type": ResultType.SINGLE,
        "reading_definitions": [{"description": "Protocol ID"}],
        "test_responses": [b"^D00518m\xae\r"]},
    "PIRI": {
        "name": "PIRI",
        "command_type": CommandType.PI18_QUERY,
        "description": "Current Settings inquiry",
        "help": " -- queries the current settings from the Inverter",
        "result_type": ResultType.COMMA_DELIMITED,
        "reading_definitions": [
            {"description": "AC Input Voltage", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/10"},
            {"description": "AC Input Current", "reading_type": ReadingType.CURRENT, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/10"},
            {"description": "AC Output Voltage", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/10"},
            {"description": "AC Output Frequency", "reading_type": ReadingType.FREQUENCY, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/10"},
            {"description": "AC Output Current", "reading_type": ReadingType.CURRENT, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/10"},
            {"description": "AC Output Apparent Power", "reading_type": ReadingType.APPARENT_POWER},
            {"description": "AC Output Active Power", "reading_type": ReadingType.WATTS},
            {"description": "Battery Voltage", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/10"},
            {"description": "Battery re-charge Voltage", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/10"},
            {"description": "Battery re-discharge Voltage", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/10"},
            {"description": "Battery Under Voltage", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/10"},
            {"description": "Battery Bulk Charge Voltage", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/10"},
            {"description": "Battery Float Charge Voltage", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/10"},
            {"description": "Battery Type",
                "reading_type": ReadingType.MESSAGE,
                "response_type": ResponseType.LIST,
                "options": BATTERY_TYPE_LIST},
            {"description": "Max AC Charging Current", "reading_type": ReadingType.CURRENT},
            {"description": "Max Charging Current", "reading_type": ReadingType.CURRENT},
            {"description": "Input Voltage Range",
                "response_type": ResponseType.LIST,
                "options": ["Appliance", "UPS"]},
            {"description": "Output Source Priority",
                "response_type": ResponseType.LIST,
                "options": ["Solar - Utility - Battery", "Solar - Battery - Utility"]},
            {"description": "Charger Source Priority",
                "response_type": ResponseType.LIST,
                "options": ["Solar First", "Solar + Utility", "Only solar charging permitted"]},
            {"description": "Max Parallel Units"},
            {"description": "Machine Type",
                "response_type": ResponseType.LIST,
                "options": ["Off Grid", "Grid Tie"]},
            {"description": "Topology",
                "response_type": ResponseType.LIST,
                "options": ["transformerless", "transformer"]},
            {"description": "Output Mode",
                "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.LIST,
                "options": OUTPUT_MODE_LIST},
            {"description": "Solar power priority",
                "response_type": ResponseType.LIST,
                "options": ["Battery-Load-Utiliy + AC Charger", "Load-Battery-Utiliy"]},
            {"description": "MPPT strings"},
            {"description": "Unknown flags?", "response_type": ResponseType.STRING},
        ],
        "test_responses": [
            b"^D0882300,217,2300,500,217,5000,5000,480,480,530,440,570,570,2,10,070,1,1,1,9,0,0,0,0,1,00\xe1k\r",
        ]},
}

COMMANDS_TO_REMOVE = []


class PI18(AbstractProtocol):
    """ pi18 protocol handler """
    def __str__(self):
        return "PI18 protocol handler"

    def __init__(self) -> None:
        super().__init__()
        self.protocol_id = b"PI18"
        self.add_command_definitions(QUERY_COMMANDS)
        self.add_command_definitions(SETTER_COMMANDS, result_type=ResultType.ACK)
        self.remove_command_definitions(COMMANDS_TO_REMOVE)
        self.check_definitions_count(expected=2)
        self.add_supported_ports([PortType.SERIAL, PortType.USB])

    def check_crc(self, response: str, command_definition: CommandDefinition = None):
        """ crc check, override for now """
        log.debug("check crc for %s in pi18", response)
        if response.startswith(b"^D"):
            # get response CRC
            data_to_check = response[:-3]
            crc_high, crc_low = crc(data_to_check)
            # print(crc_high, crc_low)
            # print(response[-3], response[-2])
            if (crc_high, crc_low) == (response[-3], response[-2]):
                return True
            else:
                log.info("PI18 response check_crc doesnt match got (%x, %x), calc (%x, %x)", crc_high, crc_low, response[-3], response[-2])
                return False
        else:
            log.info("PI18 response doesnt start with ^D - check_crc fails")
            return False
        log.info("PI18 response check_crc fall through")
        return False

    def trim_response(self, response: str, command_definition: CommandDefinition = None) -> str:
        """ Remove extra characters from response """
        log.debug("trim %s, definition: %s", response, command_definition)
        if response.startswith(b"^D"):
            # trim ^Dxxx where xxx is data length
            response = response[5:]
        if response.endswith(b'\r'):
            # has checksum, so trim last 3 chars
            response = response[:-3]
        if response.startswith(b'('):
            # pi30 style response
            response = response[1:]
        return response

    def get_full_command(self, command: str) -> bytes:
        """ generate the full command including prefix, crc and \n as needed """
        log.info("Using protocol: %s with %i commands", self.protocol_id, len(self.command_definitions))
        command_defn = self.get_command_definition(command)

        # raise exception if no command definition is found
        if command_defn is None:
            raise CommandDefinitionMissing(f"No definition found in PI18 for {command}")

        # full command is ^PlllCCCcrc\n or ^SlllCCCcrc\n
        # lll = length of all except ^Dlll
        # CCC = command
        # crc = 2 bytes
        length = len(command) + 3
        # Determine prefix
        match command_defn.command_type:
            case CommandType.PI18_QUERY:
                prefix = "^P"
            case CommandType.PI18_SETTER:
                prefix = "^S"
            case _:
                # edge case / default PI30 command / maybe this should raise an error
                prefix = "("
        full_command = bytes(f"{prefix}{length:#03d}{command}", "utf-8")
        crc_high, crc_low = crc(full_command)
        full_command += bytes([crc_high, crc_low, 13])

        log.debug("full command: %s", full_command)
        return full_command



""" powermon / ports / usbport.py """
import logging
import os
import time

from powermon.commands.command import Command
from powermon.commands.result import Result, ResultType
from powermon.dto.portDTO import PortDTO
from powermon.ports.abstractport import AbstractPort
from powermon.ports.porttype import PortType
from powermon.protocols import get_protocol_definition

log = logging.getLogger("USBPort")


class USBPort(AbstractPort):
    """ usb port object """
    @classmethod
    def from_config(cls, config=None):
        log.debug("building usb port. config:%s", config)
        path = config.get("path", "/dev/hidraw0")
        # get protocol handler, default to PI30 if not supplied
        protocol = get_protocol_definition(protocol=config.get("protocol", "PI30"))
        return cls(path=path, protocol=protocol)

    def __init__(self, path, protocol) -> None:
        super().__init__(protocol=protocol)
        self.port_type = PortType.USB
        self.path = path
        self.port = None
        self.is_protocol_supported()

    def to_dto(self):
        dto = PortDTO(type="usb", path=self.path, protocol=self.protocol.to_dto())
        return dto

    def is_connected(self) -> bool:
        return self.port is not None

    async def connect(self) -> bool:
        if self.is_connected():
            log.debug("USBPort already connected")
            return True
        log.debug("USBPort connecting. path:%s, protocol:%s", self.path, self.protocol)
        try:
            self.port = os.open(self.path, os.O_RDWR | os.O_NONBLOCK)
            log.debug("USBPort port number $%s", self.port)
        except Exception as e:
            log.warning("Error openning usb port: %s", e)
            self.port = None
            self.error_message = e
        return self.is_connected()

    async def disconnect(self) -> None:
        log.debug("USBPort disconnecting: %i", self.port)
        if self.port is not None:
            os.close(self.port)
        self.port = None

    async def send_and_receive(self, command: Command) -> Result:
        if not self.is_connected():
            log.warning("USBPort not connected")
            return command.build_result(result_type=ResultType.ERROR, raw_response=b"USBPort not connected", protocol=self.protocol)
        response_line = bytes()

        # Send the command to the open usb connection
        full_command = command.full_command
        cmd_len = len(full_command)
        log.debug("length of to_send: %i", cmd_len)
        # for command of len < 8 it ok just to send
        # otherwise need to pack to a multiple of 8 bytes and send 8 at a time
        if cmd_len <= 8:
            # Send all at once
            log.debug("sending full_command in on shot")
            time.sleep(0.05)
            os.write(self.port, full_command)
        else:
            log.debug("multiple chunk send")
            chunks = [full_command[i:i + 8] for i in range(0, cmd_len, 8)]
            for chunk in chunks:
                # pad chunk to 8 bytes
                if len(chunk) < 8:
                    padding = 8 - len(chunk)
                    chunk += b'\x00' * padding
                log.debug("sending chunk: %s", chunk)
                time.sleep(0.05)
                os.write(self.port, chunk)
        time.sleep(0.25)
        # Read from the usb connection
        # try to a max of 100 times
        for _ in range(100):
            # attempt to deal with resource busy and other failures to read
            try:
                time.sleep(0.15)
                r = os.read(self.port, 256)
                response_line += r
            except Exception as e:
                log.debug("USB read error: %s", e)
            # Finished is \r is in byte_response
            if bytes([13]) in response_line:
                # remove anything after the \r
                response_line = response_line[: response_line.find(bytes([13])) + 1]
                break
        log.debug("usb response was: %s", response_line)
        # response = self.get_protocol().check_response_and_trim(response_line)
        result = command.build_result(raw_response=response_line, protocol=self.protocol)

        return result



""" protocols / ved.py """
import logging
from struct import unpack

import construct as cs

from powermon.commands.command import CommandType
from powermon.commands.command_definition import CommandDefinition
from powermon.commands.reading_definition import ReadingType, ResponseType
from powermon.commands.result import ResultType
from powermon.errors import CommandError, InvalidCRC, InvalidResponse
from powermon.ports.porttype import PortType
from powermon.protocols.abstractprotocol import AbstractProtocol
from powermon.protocols.helpers import victron_checksum

log = logging.getLogger("daly")

soc_construct = cs.Struct(
    "start_flag" / cs.Bytes(1),
    "module_address" / cs.Bytes(1),
    "command_id" / cs.Bytes(1),
    "data_length" / cs.Byte,
    "battery_voltage" / cs.Int16ub,
    "acquistion_voltage" / cs.Int16ub,
    "current" / cs.Int16ub,
    "soc" / cs.Int16ub,
    "checksum" / cs.Bytes(1)
)

COMMANDS = {
    "SOC": {
        "name": "SOC",
        "description": "State of Charge",
        "help": " -- display the battery state of charge",
        # "type": "DALY",
        "command_type": CommandType.SERIAL_READ_UNTIL_DONE,
        "command_code": "90",
        "result_type": ResultType.CONSTRUCT,
        "construct": soc_construct,
        "reading_definitions": [
            {"index": "start_flag", "description": "start flag", "reading_type": ReadingType.HEX_STR, "response_type": ResponseType.HEX_CHAR},
            {"index": "module_address", "description": "module address", "reading_type": ReadingType.HEX_STR, "response_type": ResponseType.HEX_CHAR},
            {"index": "command_id", "description": "command id", "reading_type": ReadingType.HEX_STR, "response_type": ResponseType.HEX_CHAR},
            {"index": "data_length", "description": "data length"},
            {"index": "battery_voltage", "description": "Battery Bank Voltage", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/10"},
            {"index": "acquistion_voltage", "description": "acquistion", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/10"},  
            {"index": "current", "description": "Current", "reading_type": ReadingType.CURRENT, "response_type": ResponseType.TEMPLATE_INT, "format_template": "(r-30000)/10"},
            {"index": "soc", "description": "SOC", "reading_type": ReadingType.PERCENTAGE, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/10"},
            {"index": "checksum", "description": "checksum", "reading_type": ReadingType.HEX_STR, "response_type": ResponseType.HEX_CHAR}],
        "test_responses": [
            b"\xa5\x01\x90\x08\x02\x10\x00\x00uo\x03\xbc\xf3",
            b"\xa5\x01\x90\x08\x02\x14\x00\x00uE\x03x\x89",
            b"\xa5\x01\x90\x08\x01\t\x00\x00u\xcf\x03\n\x99",
        ],
    },
}


class Daly(AbstractProtocol):
    """
    Daly BMS protocol handler
    """

    def __str__(self):
        return "DALY protocol handler for DALY BMS"

    def __init__(self) -> None:
        super().__init__()
        self.protocol_id = b"DALY"
        self.add_command_definitions(COMMANDS)
        self.add_supported_ports([PortType.SERIAL, PortType.USB])
        self.check_definitions_count(expected=None)

    def get_full_command(self, command) -> bytes:
        """
        Override the default get_full_command
        """
        log.info("Using protocol %s with %i commands", self.protocol_id, len(self.command_definitions))

        command_definition : CommandDefinition = self.get_command_definition(command)
        if command_definition is None:
            return None

        # DALY commands
        #
        #
        # 95 -> a58095080000000000000000c2
        #       a58090080000000000000000bd
        source = 0x80  # 4 = USB, 8 = Bluetooth
        command = command_definition.command_code
        data_length = 8
        full_command = bytearray()
        full_command.append(0xa5)  # start flag
        full_command.append(source)
        full_command.append(bytes.fromhex(command_definition.command_code)[0])
        full_command.append(data_length)
        full_command += bytearray(data_length)
        full_command.append(sum(full_command) & 0xFF)
        full_command.append(10)
        full_command = bytes(full_command)
        log.debug("full_command: %s", full_command)
        return full_command

        command_type = command_definition.command_type
        match command_type:
            case CommandType.VICTRON_GET:
                # command components
                raw_command_code = command_definition.command_code  # eg 1000 for batteryCapacity
                if raw_command_code is None:
                    raise CommandError(f"command_code not found for {command=} - check protocol definition for this command")
                command_code = f"{unpack('<h', bytes.fromhex(raw_command_code))[0]:04X}"
                flags = "00"

                # build command
                cmd = f"{command_type.value}{command_code}{flags}"
                # pad cmd and convert to bytes and determine checksum
                checksum = victron_checksum(bytes.fromhex(f"0{cmd}"))

                # build full command
                cmd = f":{cmd}{checksum:02X}\n".encode()
                log.debug("full command: %s", cmd)
                return cmd
            case CommandType.VICTRON_LISTEN:
                # Just listen - dont need to send a command
                log.debug("command is LISTEN type so returning %s", command_type)
                return command_type
        raise CommandError(f"unable to generate full command for {command}, type {command_type} - is the definition wrong or CommandType not implemented?")

    def check_valid(self, response: str, command_definition: CommandDefinition = None) -> bool:
        """ check response is valid """
        log.debug("check valid for %s, definition: %s", response, command_definition)
        return True
        # if response is None:
        #     raise InvalidResponse("Response is None")
        # if len(response) <= 3:
        #     raise InvalidResponse("Response is too short")
        # command_type = command_definition.command_type
        # match command_type:
        #     case CommandType.VICTRON_GET:
        #         if response.count(b':') != 1:
        #             raise InvalidResponse("Response incomplete - missing ':'")
        # return True

    def check_crc(self, response: str, command_definition: CommandDefinition = None) -> bool:
        """ crc check, needs override in protocol """
        log.debug("checking crc for %s", response)
        return True
        # command_type = command_definition.command_type
        # match command_type:
        #     case CommandType.VICTRON_GET:
        #         # HEX protocol response
        #         log.debug("checking validity of '%s'", response)
        #         _r = response.split(b":")[1][:-1].decode()
        #         # print(f"trimmed response {_r}")
        #         _r = f"0{_r}"
        #         # print(f"padded response {_r}")
        #         _r = bytes.fromhex(_r)
        #         # print(f"bytes response {_r}")
        #         data = _r[:-1]
        #         checksum = _r[-1:][0]
        #         expected_checksum = victron_checksum(data)
        #         if expected_checksum == checksum:
        #             log.debug("VED Hex Checksum matches in response '%s' checksum:'%s'", response, checksum)
        #             return True
        #         else:
        #             # print("VED Hex Checksum does not match")
        #             raise InvalidCRC(f"response has invalid CRC - got '\\x{checksum:02x}', calculated '\\x{expected_checksum:02x}")
        #     case CommandType.VICTRON_LISTEN:
        #         return True
        # return True

    def trim_response(self, response: str, command_definition: CommandDefinition = None) -> str:
        """ Remove extra characters from response """
        log.debug("response: %s", response)
        return response
        # command_type = command_definition.command_type
        # _ret = None
        # match command_type:
        #     case CommandType.VICTRON_GET:
        #         # HEX response, e.g. b":70010007800C6\n"
        #         _ret = response.split(b":")[1][:-3]
        #     case CommandType.VICTRON_LISTEN:
        #         # VEDTEXT response, return the lot
        #         _ret = response
        # log.debug("trim_response: %s", _ret)
        # return _ret



""" powermon / ports / serialport.py """
import logging
import time

import serial
import asyncio

from powermon.commands.command import Command, CommandType
from powermon.commands.result import Result
from powermon.dto.portDTO import PortDTO
from powermon.ports.abstractport import AbstractPort
from powermon.ports.porttype import PortType
from powermon.protocols import get_protocol_definition

log = logging.getLogger("SerialPort")


class SerialPort(AbstractPort):
    """ serial port object - normally a usb to serial adapter """

    def __str__(self):
        return f"SerialPort: {self.path=}, {self.baud=}, protocol:{self.protocol}, {self.serial_port=}, {self.error_message=}"

    @classmethod
    def from_config(cls, config=None):
        log.debug("building serial port. config:%s", config)
        path = config.get("path", "/dev/ttyUSB0")
        baud = config.get("baud", 2400)
        # get protocol handler, default to PI30 if not supplied
        protocol = get_protocol_definition(protocol=config.get("protocol", "PI30"))
        return cls(path=path, baud=baud, protocol=protocol)

    def __init__(self, path, baud, protocol) -> None:
        super().__init__(protocol=protocol)
        self.port_type = PortType.SERIAL
        self.path = path
        self.baud = baud
        self.serial_port = None
        self.is_protocol_supported()
        # self.error_message = None

    def to_dto(self) -> PortDTO:
        dto = PortDTO(type="serial", path=self.path, baud=self.baud, protocol=self.protocol.to_dto())
        return dto

    def is_connected(self):
        return self.serial_port is not None and self.serial_port.is_open

    async def connect(self) -> int:
        log.debug("usbserial port connecting. path:%s, baud:%s", self.path, self.baud)
        try:
            self.serial_port = serial.Serial(port=self.path, baudrate=self.baud, timeout=1, write_timeout=1)
            log.debug(self.serial_port)
        except ValueError as e:
            log.error("Incorrect configuration for serial port: %s", e)
            self.error_message = str(e)
            self.serial_port = None
        except serial.SerialException as e:
            log.error("Error opening serial port: %s", e)
            self.error_message = str(e)
            self.serial_port = None
        return self.is_connected()

    async def disconnect(self) -> None:
        log.debug("usbserial port disconnecting")
        if self.serial_port is not None:
            self.serial_port.close()
        self.serial_port = None

    async def send_and_receive(self, command: Command) -> Result:
        full_command = command.full_command
        response_line = None
        log.info("port: %s, full_command: %s", self.serial_port, full_command)
        if not self.is_connected():
            raise RuntimeError("Serial port not open")
        try:
            log.debug("Executing command via usbserial...")
            self.serial_port.reset_input_buffer()
            self.serial_port.reset_output_buffer()
            # Process i/o differently depending on command type
            command_defn = command.command_definition
            match command_defn.command_type:
                case CommandType.VICTRON_LISTEN:
                    # this command type doesnt need to send a command, it just listens on the serial port
                    _lines = 30
                    log.debug("case: CommandType.VICTRON_LISTEN, listening for %i lines", _lines)
                    response_line = b""
                    for _ in range(_lines):
                        _response = self.serial_port.read_until(b"\n")
                        response_line += _response
                case CommandType.SERIAL_READONLY:
                    # read until no more data
                    log.debug("CommandType.SERIAL_READONLY")
                    response_line = b""
                    while True:
                        await asyncio.sleep(0.2)  # give serial port time to receive the data
                        to_read = self.serial_port.in_waiting
                        log.debug("bytes waiting: %s", to_read)
                        if to_read == 0:
                            break
                        # got some data to read
                        response_line += self.serial_port.read(to_read)
                case CommandType.SERIAL_READ_UNTIL_DONE:
                    # this case reads until no more to read or timeout
                    log.debug("case: CommandType.SERIAL_READ_UNTIL_DONE")
                    response_line = b""
                    self.serial_port.timeout = 0.5
                    self.serial_port.write_timeout = 1
                    self.serial_port.reset_input_buffer()
                    self.serial_port.reset_output_buffer()
                    c = self.serial_port.write(full_command)
                    log.debug("wrote %s bytes", c)
                    self.serial_port.flush()
                    # read until no more data
                    while True:
                        # await asyncio.sleep(0.5)  # give serial port time to receive the data
                        time.sleep(0.5)
                        to_read = self.serial_port.in_waiting
                        log.debug("bytes waiting %s", to_read)
                        if to_read == 0:
                            break
                        # got some data to read
                        response_line += self.serial_port.read(to_read)
                case _:
                    # default processing
                    c = self.serial_port.write(full_command)
                    log.debug("Default serial s&r. Wrote %i bytes", c)
                    self.serial_port.flush()
                    time.sleep(0.1)  # give serial port time to receive the data
                    response_line = self.serial_port.read_until(b"\r")
            log.info("serial response was: %s", response_line)
            # response = self.get_protocol().check_response_and_trim(response_line)
            result = command.build_result(raw_response=response_line, protocol=self.protocol)
            return result
        except Exception as e:
            log.warning("Serial read error: %s", e)
            result.error = True
            result.error_messages.append(f"Serial read error {e}")
            self.disconnect()
            return result



""" powermon / ports / __init__.py """
import logging

from powermon.errors import ConfigError
from powermon.ports.porttype import PortType
from powermon.ports.serialport import SerialPort
from powermon.ports.testport import TestPort
from powermon.ports.usbport import USBPort

# Set-up logger
log = logging.getLogger("ports")


def from_config(port_config):
    """ get a port object from config data """
    log.debug("port_config: %s", port_config)

    port_object = None
    if not port_config:
        raise ConfigError("no port config supplied")

    # port type is mandatory
    port_type = port_config.get("type")
    log.debug("portType: %s", port_type)

    # return None if port type is not defined
    if port_type is None:
        return None

    # build port object
    match port_type:
        case PortType.TEST:
            port_object = TestPort.from_config(config=port_config)
        case PortType.SERIAL:
            port_object = SerialPort.from_config(config=port_config)
        case PortType.USB:
            port_object = USBPort.from_config(config=port_config)
        # Pattern for port types that cause problems when imported
        case PortType.BLE:
            log.debug("port_type BLE found")
            from powermon.ports.bleport import BlePort
            port_object = BlePort.from_config(config=port_config)
        case _:
            log.info("port type object not found for %s", port_type)
            raise ConfigError(f"Invalid port type: '{port_type}'")

    return port_object
